# TPK4186 - Advanced Tools for Performance Engineering
# Matrices.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# 1. Required modules
# 2. Creator
# 3. Access functions
# 4. Printer
# 5. Operations
# 6. Inverse
# X. Tests

# 1. Required modules
# -------------------

import sys
import math

# 2. Creator
# ----------

def MatrixNew(numberOfRows, numberOfColumns):
  return [[0.0 for c in range(0, numberOfColumns)] for r in range(0, numberOfRows)]

# 3. Access functions
# -------------------

def MatrixGetNumberOfRows(matrix):
  return len(matrix)

def MatrixGetNumberOfColumns(matrix):
  return len(matrix[0])

def MatrixGet(matrix, rowIndex, columnIndex):
  if rowIndex<0 or rowIndex>=MatrixGetNumberOfRows(matrix):
    return None
  if columnIndex<0 or columnIndex>=MatrixGetNumberOfColumns(matrix):
    return None
  return matrix[rowIndex][columnIndex]

def MatrixSet(matrix, rowIndex, columnIndex, value):
  if rowIndex<0 or rowIndex>=MatrixGetNumberOfRows(matrix):
    return None
  if columnIndex<0 or columnIndex>=MatrixGetNumberOfColumns(matrix):
    return None
  matrix[rowIndex][columnIndex] = value

# 4. Printer
# ----------

def MatrixPrint(matrix):
  for rowIndex in range(0, MatrixGetNumberOfRows(matrix)):
    for columnIndex in range(0, MatrixGetNumberOfColumns(matrix)):
      if columnIndex>0:
        sys.stdout.write("\t")
      sys.stdout.write("%f" % MatrixGet(matrix, rowIndex, columnIndex))
    sys.stdout.write("\n")

# 5. Operations
# -------------
  
def MatrixNewUnit(numberOfRows):
  unit = MatrixNew(numberOfRows, numberOfRows)
  for index in range(0, numberOfRows):
    MatrixSet(unit, index, index, 1.0)
  return unit

def MatrixCopy(matrix):
  result = MatrixNew(MatrixGetNumberOfRows(matrix), MatrixGetNumberOfColumns(matrix))
  for rowIndex in range(0, MatrixGetNumberOfRows(result)):
    for columnIndex in range(0, MatrixGetNumberOfColumns(result)):
      v = MatrixGet(matrix, rowIndex, columnIndex)
      MatrixSet(result, rowIndex, columnIndex, v)
  return result

def MatrixTranspose(matrix):
  result = MatrixNew(MatrixGetNumberOfColumns(matrix), MatrixGetNumberOfRows(matrix))
  for rowIndex in range(0, MatrixGetNumberOfRows(matrix)):
    for columnIndex in range(0, MatrixGetNumberOfColumns(matrix)):
      v = MatrixGet(matrix, rowIndex, columnIndex)
      MatrixSet(result, columnIndex, rowIndex, v)
  return result

def MatrixMultiplyByScalar(matrix, scalar):
  result = MatrixNew(MatrixGetNumberOfRows(matrix), MatrixGetNumberOfColumns(matrix))
  for rowIndex in range(0, MatrixGetNumberOfRows(result)):
    for columnIndex in range(0, MatrixGetNumberOfColumns(result)):
      v = MatrixGet(matrix, rowIndex, columnIndex)
      MatrixSet(result, rowIndex, columnIndex, scalar*v)
  return result

def MatrixSum(matrix1, matrix2):
  if MatrixGetNumberOfRows(matrix1)!=MatrixGetNumberOfRows(matrix2):
    return None
  if MatrixGetNumberOfColumns(matrix1)!=MatrixGetNumberOfColumns(matrix2):
    return None
  result = MatrixNew(MatrixGetNumberOfRows(matrix1), MatrixGetNumberOfColumns(matrix1))
  for rowIndex in range(0, MatrixGetNumberOfRows(result)):
    for columnIndex in range(0, MatrixGetNumberOfColumns(result)):
      v1 = MatrixGet(matrix1, rowIndex, columnIndex)
      v2 = MatrixGet(matrix2, rowIndex, columnIndex)
      MatrixSet(result, rowIndex, columnIndex, v1+v2)
  return result

def MatrixProduct(matrix1, matrix2):
  if MatrixGetNumberOfColumns(matrix1)!=MatrixGetNumberOfRows(matrix2):
    return None
  result = MatrixNew(MatrixGetNumberOfRows(matrix1), MatrixGetNumberOfColumns(matrix2))
  for rowIndex in range(0, MatrixGetNumberOfRows(result)):
    for columnIndex in range(0, MatrixGetNumberOfColumns(result)):
      s = 0.0
      for index in range(0, MatrixGetNumberOfColumns(matrix1)):
        v1 = MatrixGet(matrix1, rowIndex, index)
        v2 = MatrixGet(matrix2, index, columnIndex)
        s += v1*v2
      MatrixSet(result, rowIndex, columnIndex, s)
  return result

def MatrixPower(matrix, n):
  if MatrixGetNumberOfColumns(matrix)!=MatrixGetNumberOfRows(matrix):
    return None
  if n==0:
    return MatrixNewUnit(MatrixGetNumberOfRows(matrix))
  if n==1:
    return MatrixCopy(matrix)
  k = n//2
  aux = MatrixPower(matrix, k)
  if n%2==0:
    result = MatrixProduct(aux, aux)
  else:
    aux = MatrixProduct(aux, aux)
    result = MatrixProduct(aux, matrix)
  return result

# 6. Inverse
# ----------

def MatrixSwapRow(matrix, rowIndex1, rowIndex2):
  if rowIndex1<0 or rowIndex1>=MatrixGetNumberOfRows(matrix):
    return
  if rowIndex2<0 or rowIndex2>=MatrixGetNumberOfRows(matrix):
    return
  if rowIndex1==rowIndex2:
    return
  row1 = matrix[rowIndex1]
  row2 = matrix[rowIndex2]
  matrix[rowIndex1] = row2
  matrix[rowIndex2] = row1
          
def MatrixMultiplyRowByScalar(matrix, rowIndex, scalar):
  if rowIndex<0 or rowIndex>=MatrixGetNumberOfRows(matrix):
    return
  for columnIndex in range(0, MatrixGetNumberOfColumns(matrix)):
    v = MatrixGet(matrix, rowIndex, columnIndex)
    MatrixSet(matrix, rowIndex, columnIndex, v*scalar)

def MatrixSubtractRow(matrix, rowIndex, subtractedRowIndex, subtractedRowScalar):
  if rowIndex<0 or rowIndex>=MatrixGetNumberOfRows(matrix):
    return
  if subtractedRowIndex<0 or subtractedRowIndex>=MatrixGetNumberOfRows(matrix):
    return
  for columnIndex in range(0, MatrixGetNumberOfColumns(matrix)):
    v = MatrixGet(matrix, rowIndex, columnIndex)
    w = MatrixGet(matrix, subtractedRowIndex, columnIndex)
    MatrixSet(matrix, rowIndex, columnIndex, v-w*subtractedRowScalar)

def MatrixInverse(matrix):
  if MatrixGetNumberOfColumns(matrix)!=MatrixGetNumberOfRows(matrix):
    return None
  M = MatrixCopy(matrix)
  R = MatrixNewUnit(MatrixGetNumberOfRows(M))
  # Triangularization of M
  for index in range(0, MatrixGetNumberOfRows(M)):
    c = MatrixGet(M, index, index)
    if c==0:
      for downIndex in range(index+1, MatrixGetNumberOfRows(M)):
        c = MatrixGet(M, downIndex, index)
        if c!=0:
          break
      if downIndex==MatrixGetNumberOfRows(M):
        break
      MatrixSwapRows(M, index, downIndex)
      MatrixSwapRows(R, index, downIndex)
    if c!=1.0:
      MatrixMultiplyRowByScalar(M, index, 1/c)
      MatrixMultiplyRowByScalar(R, index, 1/c)
    for downIndex in range(index+1, MatrixGetNumberOfRows(M)):
      c = MatrixGet(M, downIndex, index)
      if c!=0:
        MatrixSubtractRow(M, downIndex, index, c)
        MatrixSubtractRow(R, downIndex, index, c)
  # Diagonalization of M
  for index in range(MatrixGetNumberOfRows(M)-1, -1, -1):
    c = MatrixGet(M, index, index)
    if c==0:
      continue
    for upIndex in range(index-1, -1, -1):
      c = MatrixGet(M, upIndex, index)
      if c!=0:
        MatrixSubtractRow(M, upIndex, index, c)
        MatrixSubtractRow(R, upIndex, index, c)
  return R
        
